/*
 *
 *
 */

#include <defs.h>

/*
 * Dispatches recieved packets.  Still pretty ugly yet.
 */
int aimd_rxdispatch(struct aim_session_t *sess, struct aim_session_t **new)
{
  struct command_rx_struct *cur = NULL;

  if (sess->queue_incoming == NULL) {
    return 0;
  } else {
    for (cur = sess->queue_incoming; cur; cur = cur->next) {
      unsigned long head;

      head = aimutil_get32(cur->data);
      if ((cur->type == 0x01) && (head == 0x00000001)) { 
	if (cur->commandlen == 4) { /* conn ack */
	  printf("connection ack\n");
	} else if (aimutil_get16(cur->data+4) == 0x0001) { /* login request */

	  if (!cur->conn->type)
	    cur->conn->type = AIM_CONN_TYPE_AUTH;

	  if (cur->conn->type != AIM_CONN_TYPE_AUTH) { /* oops */
	    printf("auth request on non-auth line\n");
	    cur->handled = 1; /* skip it */
	    continue;
	  }

	  cur->handled = aim_parse_incomingauth(sess, cur);
	} else if (aimutil_get16(cur->data+4) == 0x0006) { /* cookie */
	  struct aimd_clientinfo *client;

	  client = aimd_findclient_bycookie(cur->data+8);
	  if (client) {
	    cur->conn->type = client->lastservrequested;
	    *new = client->sess;
	    return 1; /* since we're going to change sessions, we dont
		       * want to continue the loop */
	  }
	  printf("could not match cookie\n");;
	} else {
	  printf("unkown channel 1 frame\n");
	}
      } else if (cur->type != 0x02) { /* non-SNAC */
	printf("connection neg\n");
      } else { /* SNAC */
	unsigned short family, subtype;

	family = aimutil_get16(cur->data);
	subtype = aimutil_get16(cur->data+2);
	
	aimd_handlesnac(sess, cur, family, subtype);
      }
      
      cur->handled = 1;
      
    } /* for */
  } /* if */

  /* 
   * This doesn't have to be called here.  It could easily be done
   * by a seperate thread or something. It's an administrative operation,
   * and can take a while. Though the less you call it the less memory
   * you'll have :)
   */
  aim_purge_rxqueue(sess);
  
  return 0;
} /* aimd_rxdispatch */

/* 
 * Connection-handling thread.  This does lots of stuff.  
 * I don't have the patience to explain it right now.
 */
void *handleconn(void *arg)
{
  int fd, selstat, keepgoing = 1;
  struct aim_session_t *sess,*newsess;
  struct aimd_sessdata *sessdata;
  struct aim_conn_t *conn,*wconn = NULL;
  struct timeval tv;

  fd = *((int *)arg);

  if (fd < 3)
    return NULL; /* error */

  sess = malloc(sizeof(struct aim_session_t));
  aim_session_init(sess);

  sessdata = malloc(sizeof(struct aimd_sessdata));
  memset(sessdata, 0, sizeof(struct aimd_sessdata));
  pthread_mutex_init(&sessdata->txqueuelock, NULL);
  pthread_mutex_init(&sessdata->rxqueuelock, NULL);
  sess->aux_data = sessdata;

  conn = aim_newconn(sess, 0 /* type not known yet */, NULL);
  conn->fd = fd;

  TXCALLLOCKED(aim_sendconnack(sess, conn);)
  while(keepgoing) {
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    TXCALLLOCKED(wconn = aim_select(sess, &tv, &selstat);); /* needs lock?? */

    switch(selstat) {
    case -1: /* error */
      keepgoing = 0;
      break;

    case 0: /* no events pending */
      break;

    case 1: /* outgoing data pending */
      TXCALLLOCKED(aim_tx_flushqueue(sess););
      break;

    case 2: /* incoming data pending */
      if (aim_get_command(sess, wconn) >= 0) {
	if (aimd_rxdispatch(sess, &newsess) == 1) {
	  struct aim_conn_t *nextconn;
	  /*
	   * If the dispatcher returns a 1, then the connection
	   * is not a new client -- its an old client with a new
	   * connection.  It also passses a pointer to the first
	   * aim session back to us.  We merge our connection with 
	   * it and continue.  Also, we free the new session.
	   */
	  nextconn = aim_conn_getnext(newsess);
	  memcpy(nextconn, conn, sizeof(struct aim_conn_t));
	  free(sess);
	  sess = newsess;
	  conn = nextconn;

	  /*
	   * This also makes a great time to do other things.
	   *
	   * The dispatcher has taken the time to set the connection
	   * type appropriatly for the service requested.
	   *
	   */
	  switch (conn->type) {
	  case AIM_CONN_TYPE_BOS:	
	    TXCALLLOCKED(aim_sendserverready(sess, conn);)
	    break;
	  default:
	    break;
	  }
	}
      } else {
	keepgoing = 0;
      }

    default: /* invalid */
      break;
    } /* switch */
  } /* while */

  /* 
   * XXX: theres a leak here.  we need a cleanup thread that goes through
   * the client list and chews up old unused sessions 
   */

  if (wconn->type == AIM_CONN_TYPE_BOS) {
    struct aimd_clientinfo *cl;

    /* if we're closing the BOS connection, we die completely. */
    cl = aimd_remclient(sess);

    if (cl) {	
      struct aimd_clientinfo *curcl;
      pthread_mutex_lock(&clientlist_lock);
      for (curcl = clientlist; curcl; curcl = curcl->next) {
	TXCALLLOCKED(aim_sendbuddyoffgoing(curcl->sess, aim_getconn_type(curcl->sess, AIM_CONN_TYPE_BOS), cl->user->userinfo.sn););
      }
      pthread_mutex_unlock(&clientlist_lock);
    }

    aim_logoff(sess);
    free(sess);
    if (cl)
      free(cl);
  } else
    aim_conn_close(conn);
  
  return NULL;
}

/*
 * Second-stage dispatcher for SNACs.  Still ugly for now.
 */
int bleck(struct aim_session_t *, struct command_rx_struct *, ...); 
void aimd_handlesnac(struct aim_session_t *sess, struct command_rx_struct *cur, unsigned short family, unsigned short subtype) 
{
  switch (family) {
  case 0x0001: 
    switch (subtype) {
    case 0x0004:
      aimd_handle_servicereq(sess, cur);
      break;
    default:
      bleck(sess, cur);
      break;
    }
    break;
  case 0x0003:
    switch (subtype) {
    case 0x0004: /* add buddy */
      /* XXX */
      {
	struct aimd_clientinfo *curcl;
	pthread_mutex_lock(&clientlist_lock);
	for (curcl = clientlist; curcl; curcl = curcl->next) {
	  printf("sending update to %s for %s\n", sess->logininfo.screen_name, curcl->user->userinfo.sn);
	  TXCALLLOCKED(aim_sendbuddyoncoming(sess, cur->conn, &curcl->user->userinfo););
	}
	pthread_mutex_unlock(&clientlist_lock);
      }
      break;
    default:
      bleck(sess, cur);
      break;
    }
    break;
  case 0x0004:
    switch (subtype) {
    case 0x0006: /* ICBM */
      aimd_handleicbm(sess, cur);
      break;
    default:
      bleck(sess, cur);
      break;
    }
    break;
  default:
    bleck(sess, cur);
    break;
  }
  return;
}
